// std::max
#include <algorithm>

#include "mpi_collective.hpp"

void pass_mpi_collective::rank_collective(function *fun)
{
	__rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun));
	reset_bb_mark(fun);
}

void pass_mpi_collective::__rank_collective(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		if (! ((bb_data *) e->src->aux)->mark1)
		{
			__rank_collective(e->src);
		}

		for (int mpi_code = 0;
		     mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		     ++mpi_code)
		{
			((bb_data *) bb->aux)->collective_rank[mpi_code] =
			    std::max(((bb_data *) bb->aux)->collective_rank[mpi_code],
			             ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
			if (((bb_data *) bb->aux)->mpi_code == mpi_code)
			{
				((bb_data *) bb->aux)->collective_rank[mpi_code]++;
				collective_max_rank[mpi_code] =
				    std::max(collective_max_rank[mpi_code],
				             ((bb_data *) bb->aux)->collective_rank[mpi_code]);
				//printf("collective max rank for %i up to %i\n",
				//		 mpi_code, collective_max_rank[mpi_code]);
			}
		}
	}
}

void pass_mpi_collective::better_rank_collective(function *fun)
{
	size_t i;
	int next_rank = 0;

	for (i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
	{
		next_rank = __better_rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun),
		                                     (enum mpi_collective_code) i,
		                                     next_rank);
	}
	reset_bb_mark(fun);
}

int pass_mpi_collective::__better_rank_collective(basic_block bb,
    mpi_collective_code mpi_code, int rank)
{
	edge e;
	edge_iterator ei;
	int next_rank;

	next_rank = rank;
	((bb_data *) bb->aux)->mark1 = mpi_code + 1;

	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		// cache or travell
		if (((bb_data *) e->src->aux)->mark1 <= mpi_code)
		{
			next_rank =
			    std::max(next_rank,
			             __better_rank_collective(e->src, mpi_code, rank));
		}
		else
		{
			next_rank =
			    std::max(next_rank,
			             ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
		}
	}
	// set mpi code
	if (((bb_data *) bb->aux)->mpi_code == mpi_code)
	{
		next_rank++;
	}
	((bb_data *) bb->aux)->collective_rank[mpi_code] = next_rank;
	return next_rank;
}

void pass_mpi_collective::get_mpi_coll_rank(function *fun,
    int rank, int mpi_code, bitmap mpi_coll)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		if (((bb_data *) bb->aux)->mpi_code == mpi_code
		    && ((bb_data *) bb->aux)->collective_rank[mpi_code] == rank)
		{
			//printf("bb %i is coll %i of rank %i\n",
			//	bb->index, mpi_code, rank);
			bitmap_set_bit(mpi_coll, bb->index);
		}
	}
}

void pass_mpi_collective::raise_warning_mpi_rank(function *fun)
{
	int mpi_code;

	bitmap_head pds_head;
	bitmap_head pdf_head;
	bitmap_head nodes_head;

	bitmap pds = &pds_head;
	bitmap pdf = &pdf_head;
	bitmap nodes = &nodes_head;
	bitmap tmp = NULL;

	bitmap_initialize(pds, &bitmap_default_obstack);
	bitmap_initialize(pdf, &bitmap_default_obstack);
	bitmap_initialize(nodes, &bitmap_default_obstack);

	bitmap_iterator bi;
	unsigned int index;

	for (mpi_code = 0; mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	     ++mpi_code)
	{

		for (int rank = 1 ; rank < collective_max_rank[mpi_code] + 1 ; rank++)
		{
			//printf("mpi_code: %i at rank: %i\n", mpi_code, i);
			// DONE: split in 2, get mpi set & find set post dom
			get_mpi_coll_rank(fun, rank, mpi_code, nodes);
			get_post_dom_bitmap(fun, nodes, pds);
			get_post_dom_frontier(fun, pds, pdf);

			//printf("----\n");
			//debug_bitmap(nodes);
			//debug_bitmap(pds);
			//debug_bitmap(pdf);

			if (! bitmap_empty_p(pdf))
			{
				// TODO make better warning messages
				printf("Mpi Collective %s incomplete rank at bbs: ",
				       mpi_collective_name[mpi_code]);
				EXECUTE_IF_SET_IN_BITMAP(nodes, 0, index, bi)
				{
					printf("%i, ", index);
				}
				printf("\n");
			}

			while (! bitmap_empty_p(pdf))
			{
				EXECUTE_IF_SET_IN_BITMAP(pdf, 0, index, bi)
				{
					printf("Divergence from bb %i\n", index);
				}

				bitmap_clear(nodes);
				bitmap_clear(pds);

				// swap nodes & pdf
				{
					tmp = nodes;
					nodes = pdf;
					pdf = tmp;
				}
				get_post_dom_bitmap(fun, nodes, pds);
				get_post_dom_frontier(fun, pds, pdf);

				//printf("----\n");
				//debug_bitmap(nodes);
				//debug_bitmap(pds);
				//debug_bitmap(pdf);
			}
			bitmap_clear(nodes);
			bitmap_clear(pds);
			bitmap_clear(pdf);
		}
	}
}

basic_block pass_mpi_collective::get_bb_from_index(function *fun, int index)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		if (bb->index == index)
		{
			return bb;
		}
	}

	return NULL;
}

void pass_mpi_collective::raise_warning_mpi_order(function *fun)
{
	int mpi_code;
	int rank;

	bitmap_head nodes_head;
	bitmap nodes = &nodes_head;
	bitmap_initialize(nodes, &bitmap_default_obstack);

	int cursor_code;
	bitmap_iterator bi;
	unsigned int index;

	int curr_coll_rank;
	bool coll_rank_diff;
	int other_coll_rank;

	for (mpi_code = 0; mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	     ++mpi_code)
	{
		for (rank = 1; rank < collective_max_rank[mpi_code] + 1 ; rank++)
		{
			//printf("mpi_code: %i at rank: %i\n", mpi_code, i);
			// DONE: split in 2, get mpi set & find set post dom
			get_mpi_coll_rank(fun, rank, mpi_code, nodes);
			//debug_bitmap(nodes);

			for (cursor_code = 0;
			     cursor_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
			     cursor_code++)
			{
				if (mpi_code == cursor_code)
				{
					continue;
				}

				curr_coll_rank = -1;
				coll_rank_diff = false;

				EXECUTE_IF_SET_IN_BITMAP(nodes, 0, index, bi)
				{
					other_coll_rank =
					    ((bb_data *) get_bb_from_index(fun, index)->aux)
					    ->collective_rank[cursor_code];
					//printf("bb %i as mpi code %i and mpi rank %i\n",
					//		index, cursor_code, other_coll_rank);
					if (curr_coll_rank == -1)
					{
						curr_coll_rank = other_coll_rank;
					}
					else if (curr_coll_rank != other_coll_rank)
					{
						coll_rank_diff = true;
					}
				}

				if (coll_rank_diff)
				{
					// TODO make better warning messages
					printf("======> WARNING !!! ====> MPI "
					       "Colelctive rank missmatch for collective %s"
					       " at MPI Collective %s at rank %i\n",
					       mpi_collective_name[cursor_code],
					       mpi_collective_name[mpi_code], rank);
				}
			}

			bitmap_clear(nodes);
		}
	}
}
